use std::{env, fs, io::stdin, process::exit};

use owo_colors::OwoColorize;

mod grid;
use grid::*;

#[derive(Debug)]
struct Machine {
	grid: Grid<Tile>,
	marbles: Vec<Pos>,

	input: Vec<u8>,
	input_index: usize,
	output: Vec<u8>,
	state: State,
	steps: usize,
}

#[derive(Debug, Default, Clone, Copy)]
enum Tile {
	#[default]
	Blank,
	Comment(u8),
	Bag,
	Marble {
		value: u8,
		dir: Direction,
	},
	Trigger,
	Digit(u8),
	Wire(WireType, bool),
	Gate(GateType),
	Print,
	Input,
	Flip,
	Math(MathOp),
	Mirror(MirrorType),
	Arrow(Direction),
}

#[derive(Debug, Clone, Copy)]
enum MirrorType {
	Forward,
	Back,
}

#[derive(Debug, Clone, Copy)]
enum MathOp {
	Add,
	Sub,
	Mul,
	Div,
	Rem,
}

#[derive(Debug, Clone, Copy)]
enum GateType {
	LessThan,
	GreaterThan,
	Equal,
	NotEqual,
}

#[derive(Debug, Clone, Copy)]
enum WireType {
	Vertical,
	Horizontal,
	Cross,
}

#[derive(Debug, Clone, Copy, PartialEq)]
enum Direction {
	Up,
	Down,
	Left,
	Right,
}

#[derive(Debug, Default, PartialEq)]
enum State {
	#[default]
	Running,
	EndOfProgram,
}

fn main() {
	let args: Vec<_> = env::args().collect();
	if args.len() <= 1 {
		println!("usage: marble source_file [input_file]");
		exit(0);
	}
	let filename = &args[1];
	let source = fs::read_to_string(filename).unwrap_or_else(|err| {
		println!("Error reading file: {err}");
		exit(1);
	});
	let input_data = args
		.get(2)
		.map(|path| {
			fs::read(path).unwrap_or_else(|err| {
				println!("Error reading file: {err}");
				exit(1);
			})
		})
		.unwrap_or_default();

	let mut interpreter = Machine::new(&source, input_data);
	print!("\x1B[2J"); // clear screen

	loop {
		interpreter.show();
		let mut input_text = String::new();
		stdin().read_line(&mut input_text).unwrap();
		let action: Vec<_> = input_text.trim().split_ascii_whitespace().collect();
		if input_text.starts_with("input ") {
			interpreter.input.extend(
				&input_text
					.strip_suffix('\n')
					.unwrap_or(&input_text)
					.as_bytes()[6..],
			);
			continue;
		}
		match action.as_slice() {
			[] | ["step"] => interpreter.step_once(),
			["dbg" | "debug"] => {
				dbg!(&interpreter);
			}
			["step", num] => _ = num.parse().map(|n| interpreter.step(n)),
			["run"] => interpreter.run(),
			["q" | "exit" | "quit"] => break,

			_ => println!("{}", "unrecognised command".red()),
		}
	}
}

impl Machine {
	fn new(source: &str, input: Vec<u8>) -> Self {
		let (rows, marbles) = parse(source);
		let grid = Grid::new(rows);
		Self {
			grid,
			marbles,
			input,
			input_index: 0,
			output: Vec::new(),
			state: State::default(),
			steps: 0,
		}
	}

	fn run(&mut self) {
		while self.state == State::Running {
			self.step_once();
			self.show();
		}
	}

	fn step(&mut self, n: usize) {
		for _ in 0..n {
			if self.state != State::Running {
				break;
			}
			self.step_once();
		}
	}

	fn show(&self) {
		// dbg!(&self);
		print!("\x1B[2J"); // clear screen
		print!("\x1B[u"); // reset cursor
		for y in 0..self.grid.height() {
			let mut marbles_on_row = Vec::new();
			for x in 0..self.grid.width() {
				let tile = self.grid.get((x, y).into());
				if let Tile::Marble { value, dir } = tile {
					marbles_on_row.push((value, dir));
					print!("{}", 'o'.magenta());
					continue;
				}
				if let Tile::Comment(c) = tile {
					print!("{}", (c as char).yellow());
					continue;
				}
				let mut powered = false;
				let c = match tile {
					Tile::Marble { value: _, dir: _ } => unreachable!(),
					Tile::Comment(_) => unreachable!(),
					Tile::Blank => ' ',
					Tile::Bag => 'B',
					Tile::Trigger => '*',
					Tile::Wire(w, state) => {
						powered = state;
						match w {
							WireType::Vertical => '|',
							WireType::Horizontal => '-',
							WireType::Cross => '+',
						}
					}
					Tile::Gate(g) => match g {
						GateType::LessThan => 'L',
						GateType::GreaterThan => 'G',
						GateType::Equal => '=',
						GateType::NotEqual => '!',
					},
					Tile::Print => 'P',
					Tile::Input => 'I',
					Tile::Flip => 'F',
					Tile::Math(m) => match m {
						MathOp::Add => 'A',
						MathOp::Sub => 'S',
						MathOp::Mul => 'M',
						MathOp::Div => 'D',
						MathOp::Rem => 'R',
					},
					Tile::Mirror(m) => match m {
						MirrorType::Forward => '/',
						MirrorType::Back => '\\',
					},
					Tile::Arrow(d) => d.char(),
					Tile::Digit(d) => d as char,
				};
				if powered {
					print!("{}", c.red());
				} else {
					print!("{}", c.cyan());
				}
			}
			print!(" :: ");
			for (val, dir) in marbles_on_row {
				print!("{}", format!("{}{:<3} ", dir.char(), val).magenta())
			}
			println!();
		}
		println!("output: {}", String::from_utf8_lossy(&self.output));
		println!("input: {}", String::from_utf8_lossy(&self.input));
		println!("steps: {}", self.steps);
	}

	fn step_once(&mut self) {
		self.steps += 1;
		// reset wires
		for y in 0..self.grid.height() {
			for x in 0..self.grid.width() {
				if let Tile::Wire(_, state) = self.grid.get_mut((x, y).into()) {
					if *state {
						*state = false;
					}
				}
			}
		}
		let mut to_remove = Vec::new();
		let mut triggers = Vec::new();
		for i in 0..self.marbles.len() {
			let marble_pos = self.marbles[i];
			let tile = self.grid.get(marble_pos);

			if let Tile::Marble { value, dir } = tile {
				let next_pos = dir.step(marble_pos);
				if !self.grid.in_bounds(next_pos) {
					continue;
				}
				let mut new_tile = None;
				let target = self.grid.get_mut(next_pos);
				match target {
					Tile::Blank => {
						*target = tile;
						self.marbles[i] = next_pos;
						new_tile = Some(Tile::Blank);
					}
					Tile::Digit(d) => {
						let new_val = value.wrapping_mul(10).wrapping_add(*d - b'0');
						*target = Tile::Marble {
							value: new_val,
							dir,
						};
						self.marbles[i] = next_pos;
						new_tile = Some(Tile::Blank);
					}
					Tile::Bag => {
						to_remove.push(i);
						new_tile = Some(Tile::Blank);
					}
					Tile::Marble {
						value: _other_value,
						dir: other_dir,
					} => {
						// bounce off other marbles
						if *other_dir != dir {
							new_tile = Some(Tile::Marble {
								value,
								dir: dir.opposite(),
							});
							*other_dir = dir;
						}
					}
					Tile::Trigger => {
						triggers.push(next_pos);
						let far_pos = dir.step(next_pos);
						let far_target = self.grid.get_mut(far_pos);
						if let Tile::Blank = far_target {
							*far_target = Tile::Marble { value, dir };
							self.marbles[i] = far_pos;
							new_tile = Some(Tile::Blank);
						}
					}
					Tile::Arrow(arrow_dir) => {
						let far_pos = arrow_dir.step(next_pos);
						let arrow_dir = *arrow_dir;
						let far_target = self.grid.get_mut(far_pos);
						if let Tile::Blank = far_target {
							self.marbles[i] = far_pos;
							*far_target = Tile::Marble {
								value,
								dir: arrow_dir,
							};
							new_tile = Some(Tile::Blank);
						} else if far_pos == marble_pos {
							// bounce on reverse arrow
							new_tile = Some(Tile::Marble {
								value,
								dir: dir.opposite(),
							});
						}
					}
					Tile::Mirror(mirror) => {
						let new_dir = mirror.new_dir(dir);
						let far_pos = new_dir.step(next_pos);
						let far_target = self.grid.get_mut(far_pos);
						if let Tile::Blank = far_target {
							*far_target = Tile::Marble {
								value,
								dir: new_dir,
							};
							self.marbles[i] = far_pos;
							new_tile = Some(Tile::Blank);
						}
					}
					_ => (),
				}

				if let Some(t) = new_tile {
					*self.grid.get_mut(marble_pos) = t;
				}
			}
		}
		let mut offset = 0;
		for i in to_remove {
			self.marbles.remove(i - offset);
			offset += 1;
		}
		for pos in triggers {
			for dir in Direction::ALL {
				self.propagate_power(dir, dir.step(pos));
			}
		}
	}

	fn propagate_power(&mut self, dir: Direction, pos: Pos) {
		if !self.grid.in_bounds(pos) {
			return;
		}
		let tile = self.grid.get_mut(pos);
		let front_pos = dir.step(pos);
		match tile {
			Tile::Wire(wiretype, state) => {
				if *state {
					return;
				}
				*state = true;
				let dirs = wiretype.directions();
				for d in dirs {
					self.propagate_power(*d, d.step(pos));
				}
			}
			Tile::Print => {
				let sample = self.grid.get(front_pos);
				if let Tile::Marble { value, dir: _ } = sample {
					self.output.push(value);
				}
			}
			Tile::Bag => {
				if self.grid.get(front_pos).is_blank() {
					*self.grid.get_mut(front_pos) = Tile::Marble { value: 0, dir };
					self.marbles.push(front_pos);
				}
			}
			Tile::Input => {
				if self.input_index < self.input.len() && self.grid.get(front_pos).is_blank() {
					let value = self.input[self.input_index];
					*self.grid.get_mut(front_pos) = Tile::Marble { value, dir };
					self.marbles.push(front_pos);
					self.input_index += 1;
				}
			}
			Tile::Math(op) => {
				let op = *op;
				let pos_a = dir.left().step(pos);
				let pos_b = dir.right().step(pos);
				let val_a = self.grid.get(pos_a).read_value();
				let val_b = self.grid.get(pos_b).read_value();
				if (!self.grid.get(pos_a).is_blank() || !self.grid.get(pos_b).is_blank())
					&& self.grid.get(front_pos).is_blank()
				{
					let result = match op {
						MathOp::Add => val_a.wrapping_add(val_b),
						MathOp::Sub => val_a.wrapping_sub(val_b),
						MathOp::Mul => val_a.wrapping_mul(val_b),
						MathOp::Div => val_a.checked_div(val_b).unwrap_or_default(),
						MathOp::Rem => val_a.checked_rem(val_b).unwrap_or_default(),
					};
					println!("{op:?} a:{val_a} b:{val_b}");
					*self.grid.get_mut(front_pos) = Tile::Marble { value: result, dir };
					self.marbles.push(front_pos);
				}
			}
			Tile::Flip => {
				let m = self.grid.get_mut(front_pos);
				match m {
					Tile::Wire(wire_type, _) => {
						*wire_type = match *wire_type {
							WireType::Vertical => WireType::Horizontal,
							WireType::Horizontal => WireType::Vertical,
							WireType::Cross => WireType::Cross,
						};
					}
					Tile::Mirror(mirror) => {
						*mirror = match *mirror {
							MirrorType::Forward => MirrorType::Back,
							MirrorType::Back => MirrorType::Forward,
						};
					}
					Tile::Arrow(dir) => {
						*dir = dir.opposite();
					}
					_ => (),
				};
			}
			Tile::Gate(gate) => {
				let gate = *gate;
				let pos_a = dir.left().step(pos);
				let pos_b = dir.right().step(pos);
				let val_a = self.grid.get(pos_a).read_value();
				let val_b = self.grid.get(pos_b).read_value();

				let result = match gate {
					GateType::LessThan => val_a < val_b,
					GateType::GreaterThan => val_a > val_b,
					GateType::Equal => val_a == val_b,
					GateType::NotEqual => val_a != val_b,
				};
				if result {
					self.propagate_power(dir, dir.step(pos));
				}
			}
			Tile::Marble { value: _, dir: _ }
			| Tile::Trigger
			| Tile::Digit(_)
			| Tile::Mirror(_)
			| Tile::Arrow(_)
			| Tile::Comment(_)
			| Tile::Blank => (),
		}
	}
}

impl Tile {
	fn read_value(&self) -> u8 {
		if let Tile::Marble { value, dir: _ } = self {
			*value
		} else {
			0
		}
	}

	fn is_blank(&self) -> bool {
		matches!(self, Tile::Blank)
	}
}

impl Direction {
	const ALL: [Direction; 4] = [
		Direction::Up,
		Direction::Down,
		Direction::Left,
		Direction::Right,
	];

	fn char(&self) -> char {
		match self {
			Direction::Up => '^',
			Direction::Down => 'v',
			Direction::Left => '<',
			Direction::Right => '>',
		}
	}

	fn opposite(&self) -> Direction {
		match self {
			Direction::Up => Direction::Down,
			Direction::Down => Direction::Up,
			Direction::Left => Direction::Right,
			Direction::Right => Direction::Left,
		}
	}

	fn right(&self) -> Direction {
		match self {
			Direction::Up => Direction::Right,
			Direction::Down => Direction::Left,
			Direction::Left => Direction::Up,
			Direction::Right => Direction::Down,
		}
	}

	fn left(&self) -> Direction {
		self.right().opposite()
	}

	fn step(&self, mut pos: Pos) -> Pos {
		match self {
			Direction::Up => pos.y -= 1,
			Direction::Down => pos.y += 1,
			Direction::Left => pos.x -= 1,
			Direction::Right => pos.x += 1,
		}
		pos
	}
}

impl WireType {
	fn directions(self) -> &'static [Direction] {
		match self {
			WireType::Vertical => &[Direction::Up, Direction::Down],
			WireType::Horizontal => &[Direction::Left, Direction::Right],
			WireType::Cross => &Direction::ALL,
		}
	}
}

impl MirrorType {
	fn new_dir(self, dir: Direction) -> Direction {
		match self {
			MirrorType::Forward => match dir {
				Direction::Up => Direction::Right,
				Direction::Down => Direction::Left,
				Direction::Left => Direction::Down,
				Direction::Right => Direction::Up,
			},
			MirrorType::Back => match dir {
				Direction::Up => Direction::Left,
				Direction::Down => Direction::Right,
				Direction::Left => Direction::Up,
				Direction::Right => Direction::Down,
			},
		}
	}
}

fn parse(source: &str) -> (Vec<Vec<Tile>>, Vec<Pos>) {
	let mut grid = Vec::new();
	let mut marbles = Vec::new();

	let mut width = 0;
	for (y, line) in source.lines().enumerate() {
		width = width.max(line.len());
		let mut tiles = Vec::new();
		let mut in_comment = false;
		for (x, char) in line.chars().enumerate() {
			if in_comment {
				if char == ')' {
					in_comment = false;
				}
				if char == ' ' {
					// allow marbles to pass through gaps in comments
					tiles.push(Tile::Blank);
				} else {
					tiles.push(Tile::Comment(char as u8));
				}
				continue;
			}
			tiles.push(match char {
				'o' => {
					marbles.push((x, y).into());
					Tile::Marble {
						value: 0,
						dir: Direction::Down,
					}
				}
				'*' => Tile::Trigger,
				'-' => Tile::Wire(WireType::Horizontal, false),
				'|' => Tile::Wire(WireType::Vertical, false),
				'+' => Tile::Wire(WireType::Cross, false),
				'/' => Tile::Mirror(MirrorType::Forward),
				'\\' => Tile::Mirror(MirrorType::Back),
				'^' => Tile::Arrow(Direction::Up),
				'v' => Tile::Arrow(Direction::Down),
				'<' => Tile::Arrow(Direction::Left),
				'>' => Tile::Arrow(Direction::Right),
				'=' => Tile::Gate(GateType::Equal),
				'!' => Tile::Gate(GateType::NotEqual),
				'L' => Tile::Gate(GateType::LessThan),
				'G' => Tile::Gate(GateType::GreaterThan),
				'P' => Tile::Print,
				'I' => Tile::Input,
				'F' => Tile::Flip,
				'A' => Tile::Math(MathOp::Add),
				'S' => Tile::Math(MathOp::Sub),
				'M' => Tile::Math(MathOp::Mul),
				'D' => Tile::Math(MathOp::Div),
				'R' => Tile::Math(MathOp::Rem),
				'B' => Tile::Bag,
				d @ '0'..='9' => Tile::Digit(d as u8),
				'#' => Tile::Comment(b'#'),
				' ' => Tile::Blank,
				'(' => {
					in_comment = true;
					Tile::Comment(b'(')
				}
				_ => Tile::Blank,
			});
		}
		grid.push(tiles);
	}
	for line in &mut grid {
		line.resize(width, Tile::Blank);
	}

	(grid, marbles)
}
